from django.apps import AppConfig
from django.conf import settings

MODULE_NAME = "policy"

DEFAULT_CFG = {
    "gql_query_policies_perms": ['101201'],
    "gql_query_policy_officers_perms": [],
    "gql_query_policies_by_insuree_perms": ['101201'],
    "gql_query_policies_by_family_perms": ['101201'],
    "gql_query_eligibilities_perms": ['101201'],
    "gql_mutation_create_policies_perms": ['101202'],
    "gql_mutation_renew_policies_perms": ['101205'],
    "gql_mutation_edit_policies_perms": ['101203'],
    "gql_mutation_suspend_policies_perms": ['101203'],
    "gql_mutation_delete_policies_perms": ['101204'],
    "policy_renewal_interval": 14,  # Notify renewal nb of days before expiry date
    "policy_location_via": "family",  # ... or product
    "default_eligibility_disabled": False,
    "activation_option": 1,
    "ACTIVATION_OPTION_CONTRIBUTION": 1,
    "CTIVATION_OPTION_PAYMENT": 2,
    "ACTIVATION_OPTION_READY": 3,
    "contribution_receipt_length": 5
}


class PolicyConfig(AppConfig):
    name = MODULE_NAME

    gql_query_policies_perms = []
    gql_query_policy_officers_perms = []
    gql_query_policies_by_insuree_perms = []
    gql_query_policies_by_family_perms = []
    gql_query_eligibilities_perms = []
    gql_mutation_create_policies_perms = []
    gql_mutation_renew_policies_perms = []
    gql_mutation_edit_policies_perms = []
    gql_mutation_suspend_policies_perms = []
    gql_mutation_delete_policies_perms = []
    policy_renewal_interval = None
    policy_location_via = None
    default_eligibility_disabled = None
    ACTIVATION_OPTION_CONTRIBUTION = None
    ACTIVATION_OPTION_PAYMENT = None
    ACTIVATION_OPTION_READY = None
    activation_option = None
    contribution_receipt_length = None

    def __load_config(self, cfg):
        for field in cfg:
            if hasattr(PolicyConfig, field):
                setattr(PolicyConfig, field, cfg[field])

    def ready(self):
        from core.models import ModuleConfiguration
        cfg = ModuleConfiguration.get_or_default(MODULE_NAME, DEFAULT_CFG)
        self.__load_config(cfg)



from policy.reports import policy_renewals
from policy.reports.policy_renewals import policy_renewals_query
from policy.reports import primary_operational_indicators
from policy.reports.primary_operational_indicators import policies_primary_indicators_query

report_definitions = [
    {
        "name": "policy_renewals",
        "engine": 0,
        "default_report": policy_renewals.template,
        "description": "Policy renewals",
        "module": "policy",
        "python_query": policy_renewals_query,
        "permission": ["131217"],
    },
    {
        "name": "policy_primary_operational_indicators",
        "engine": 0,
        "default_report": primary_operational_indicators.template,
        "description": "Policy primary operational indicators",
        "module": "policy",
        "python_query": policies_primary_indicators_query,
        "permission": ["131201"],
    },
]



# Generated by Django 3.0.3 on 2020-06-16 10:20

import core.fields
import datetime
from django.db import migrations, models
import uuid


class Migration(migrations.Migration):

    dependencies = [
        ('policy', '0001_initial'),
    ]

    operations = [
        migrations.CreateModel(
            name='PolicyRenewal',
            fields=[
                ('validity_from', core.fields.DateTimeField(db_column='ValidityFrom', default=datetime.datetime.now)),
                ('validity_to', core.fields.DateTimeField(blank=True, db_column='ValidityTo', null=True)),
                ('legacy_id', models.IntegerField(blank=True, db_column='LegacyID', null=True)),
                ('id', models.AutoField(db_column='RenewalID', primary_key=True, serialize=False)),
                ('uuid', models.CharField(db_column='RenewalUUID', default=uuid.uuid4, max_length=36, unique=True)),
                ('renewal_prompt_date', core.fields.DateField(db_column='RenewalPromptDate')),
                ('renewal_date', core.fields.DateField(db_column='RenewalDate')),
                ('phone_number', models.CharField(blank=True, db_column='PhoneNumber', max_length=25, null=True)),
                ('sms_status', models.SmallIntegerField(db_column='SMSStatus', default=0)),
                ('renewal_warnings', models.SmallIntegerField(blank=True, db_column='RenewalWarnings', default=0, null=True)),
                ('response_status', models.IntegerField(blank=True, db_column='ResponseStatus', default=0, null=True)),
                ('response_date', core.fields.DateTimeField(blank=True, db_column='ResponseDate', null=True)),
                ('audit_user_id', models.IntegerField(blank=True, db_column='AuditCreateUser', null=True)),
            ],
            options={
                'db_table': 'tblPolicyRenewals',
                'managed': False,
            },
        ),
        migrations.CreateModel(
            name='PolicyRenewalDetail',
            fields=[
                ('validity_from', core.fields.DateTimeField(db_column='ValidityFrom', default=datetime.datetime.now)),
                ('validity_to', core.fields.DateTimeField(blank=True, db_column='ValidityTo', null=True)),
                ('legacy_id', models.IntegerField(blank=True, db_column='LegacyID', null=True)),
                ('id', models.AutoField(db_column='RenewalDetailID', primary_key=True, serialize=False)),
                ('audit_user_id', models.IntegerField(blank=True, db_column='AuditCreateUser', null=True)),
            ],
            options={
                'db_table': 'tblPolicyRenewalDetails',
                'managed': False,
            },
        ),
    ]



# Generated by Django 2.1.8 on 2019-04-05 08:15

import core.fields
from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='Policy',
            fields=[
                ('id', models.AutoField(db_column='PolicyID', primary_key=True, serialize=False)),
                ('legacy_id', models.IntegerField(blank=True, db_column='LegacyID', null=True)),
                ('stage', models.CharField(blank=True, db_column='PolicyStage', max_length=1, null=True)),
                ('status', models.SmallIntegerField(blank=True, db_column='PolicyStatus', null=True)),
                ('value', models.DecimalField(blank=True, db_column='PolicyValue', decimal_places=2, max_digits=18, null=True)),
                ('enroll_date', core.fields.DateField(db_column='EnrollDate')),
                ('start_date', core.fields.DateField(db_column='StartDate')),
                ('effective_date', core.fields.DateField(blank=True, db_column='EffectiveDate', null=True)),
                ('expiry_date', core.fields.DateField(blank=True, db_column='ExpiryDate', null=True)),
                ('validity_from', core.fields.DateTimeField(db_column='ValidityFrom')),
                ('validity_to', core.fields.DateTimeField(blank=True, db_column='ValidityTo', null=True)),
                ('offline', models.BooleanField(blank=True, db_column='isOffline', null=True)),
                ('audit_user_id', models.IntegerField(db_column='AuditUserID')),
                ('row_id', models.BinaryField(blank=True, db_column='RowID', null=True)),
            ],
            options={
                'db_table': 'tblPolicy',
                'managed': False,
            },
        ),
    ]



import base64
import json
from dataclasses import dataclass
from core.utils import filter_validity
from core.models import User
from core.test_helpers import create_test_interactive_user
from core.models.openimis_graphql_test_case import openIMISGraphQLTestCase

from django.conf import settings
from medical.models import Service 
from graphene_django.utils.testing import GraphQLTestCase
from graphql_jwt.shortcuts import get_token
#credits https://docs.graphene-python.org/projects/django/en/latest/testing/
from medical.test_helpers import (
    create_test_item,
    create_test_service
)
from insuree.test_helpers import create_test_insuree
from policy.test_helpers import create_test_policy, dts
from contribution.test_helpers import create_test_premium
from product.models import ProductItemOrService
from product.test_helpers import (
    create_test_product,
    create_test_product_service,
    create_test_product_item,
)
from location.test_helpers import (
    create_test_health_facility,
    create_test_village
)
from uuid import UUID
@dataclass
class DummyContext:
    """ Just because we need a context to generate. """
    user: User

class PolicyGraphQLTestCase(openIMISGraphQLTestCase):

    admin_user = None
  
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.admin_user = create_test_interactive_user(username="testLocationAdmin")
        cls.admin_token = get_token(cls.admin_user, DummyContext(user=cls.admin_user))
        
        cls.test_village  =create_test_village()
        cls.test_ward =cls.test_village.parent
        cls.test_region =cls.test_village.parent.parent.parent
        cls.test_district = cls.test_village.parent.parent
        # Given
        cls.insuree = create_test_insuree(custom_props={'current_village':cls.test_village})
        
        cls.service = create_test_service("A", custom_props={"name": "test_simple_batch"})
        cls.item = create_test_item("A", custom_props={"name": "test_simple_batch"})
        
        cls.product = create_test_product(
            "BCUL0001",
            custom_props={
                "name": "simplebatch",
                "lump_sum": 10_000,
                "location_id": cls.test_region.id
            },
        )
        
        cls.product_service = create_test_product_service(
            cls.product,
            cls.service,
            custom_props={},
        )
        cls.product_item = create_test_product_item(
            cls.product,
            cls.item,
            custom_props={"price_origin": ProductItemOrService.ORIGIN_RELATIVE},
        )
        cls.policy = create_test_policy(cls.product, cls.insuree, link=True)
        cls.premium = create_test_premium(
            policy_id=cls.policy.id, custom_props={}
        )
        cls.policy_past = create_test_policy(cls.product, cls.insuree, link=True, custom_props={
            "enroll_date": dts("2010-01-01"),
            "start_date": dts("2010-01-01"),
            "validity_from": dts("2010-01-01"),
            "effective_date": dts("2010-01-01"),
            "expiry_date": dts("2011-01-01"),
            })
        cls.premium_past = create_test_premium(
            policy_id=cls.policy_past.id, custom_props={'pay_date':dts('2010-01-01')}
        )
        cls.not_insuree = create_test_insuree(with_family= False, custom_props={'family':cls.insuree.family})

            
            
    def test_insuree_policy_query(self):
        
        response = self.query(
            '''
            query {
                policies(first: 10,orderBy: ["-enrollDate"], balanceLte: 100)
                {
                    totalCount
                    pageInfo { hasNextPage, hasPreviousPage, startCursor, endCursor}
                    edges
                    {
                        node
                        {
                            uuid,product{id,code,name,location{id}},officer{id,uuid,code,lastName,otherNames},family{id,uuid,headInsuree{id chfId uuid lastName otherNames},location{id,uuid,code,name,type,parent{id,uuid,code,name,type,parent{id,uuid,code,name,type,parent{id,uuid,code,name,type}}}}},enrollDate,effectiveDate,startDate,expiryDate,stage,status,value,sumPremiums,validityFrom,validityTo
                        }
                    }
                }
            }
            ''',
            headers={"HTTP_AUTHORIZATION": f"Bearer {self.admin_token}"},
        )

        content = json.loads(response.content)

        # This validates the status code and if you get errors
        self.assertResponseNoErrors(response)

        # Add some more asserts if you like
        ...
    def test_query_not_insured_family_member(self):
        response = self.query(
            '''
    
            query policiesByInsuree($chfid: String!) {
                policiesByInsuree(chfId:$chfid)
                {
                    totalCount
                    pageInfo { hasNextPage, hasPreviousPage, startCursor, endCursor}
                    edges
                    {
                        node
                        {
                            policyUuid,productCode,productName,officerCode,officerName,enrollDate,effectiveDate,startDate,expiryDate,status,policyValue,balance,ded,dedInPatient,dedOutPatient,ceiling,ceilingInPatient,ceilingOutPatient
                        }
                    }
                }
            }
            ''',
            headers={"HTTP_AUTHORIZATION": f"Bearer {self.admin_token}"},
            variables={'chfid': self.not_insuree.chf_id, 'targetDate':self.policy.effective_date.strftime("%Y-%m-%d")}
        )

        content = json.loads(response.content)

        # This validates the status code and if you get errors
        self.assertResponseNoErrors(response)
        self.assertEqual(content['data']['policiesByInsuree']['totalCount'], 0)

    def test_query_with_variables(self):
        response = self.query(
            '''
    
            query policiesByInsuree($chfid: String!) {
                policiesByInsuree(chfId:$chfid)
                {
                    totalCount
                    pageInfo { hasNextPage, hasPreviousPage, startCursor, endCursor}
                    edges
                    {
                        node
                        {
                            policyUuid,productCode,productName,officerCode,officerName,enrollDate,effectiveDate,startDate,expiryDate,status,policyValue,balance,ded,dedInPatient,dedOutPatient,ceiling,ceilingInPatient,ceilingOutPatient
                        }
                    }
                }
            }
            ''',
            headers={"HTTP_AUTHORIZATION": f"Bearer {self.admin_token}"},
            variables={'chfid': self.insuree.chf_id, 'targetDate':self.policy.effective_date.strftime("%Y-%m-%d")}
        )

        content = json.loads(response.content)

        # This validates the status code and if you get errors
        self.assertResponseNoErrors(response)
        self.assertEqual(content['data']['policiesByInsuree']['totalCount'], 2)
        
    def test_query_with_variables_2(self):
        response = self.query(
            '''
    
            query policiesByInsuree($chfid: String!, $activeOrLastExpiredOnly: Boolean!) {
                policiesByInsuree(chfId:$chfid, activeOrLastExpiredOnly:$activeOrLastExpiredOnly)
                {
                    totalCount
                    pageInfo { hasNextPage, hasPreviousPage, startCursor, endCursor}
                    edges
                    {
                        node
                        {
                            policyUuid,productCode,productName,officerCode,officerName,enrollDate,effectiveDate,startDate,expiryDate,status,policyValue,balance,ded,dedInPatient,dedOutPatient,ceiling,ceilingInPatient,ceilingOutPatient
                        }
                    }
                }
            }
            ''',
            headers={"HTTP_AUTHORIZATION": f"Bearer {self.admin_token}"},
            variables={'chfid': self.insuree.chf_id, 'activeOrLastExpiredOnly':True}
        )

        content = json.loads(response.content)

        # This validates the status code and if you get errors
        self.assertResponseNoErrors(response)
        self.assertEqual(content['data']['policiesByInsuree']['totalCount'], 1)
        self.assertEqual(UUID(content['data']['policiesByInsuree']['edges'][0]['node']['policyUuid']), UUID(self.policy.uuid))
                
    def test_query_with_variables_3(self):
        response = self.query(
            '''
    
            query policiesByInsuree($chfid: String!, $targetDate:  Date! ) {
                policiesByInsuree(chfId:$chfid ,targetDate: $targetDate)
                {
                    totalCount
                    pageInfo { hasNextPage, hasPreviousPage, startCursor, endCursor}
                    edges
                    {
                        node
                        {
                            policyUuid,productCode,productName,officerCode,officerName,enrollDate,effectiveDate,startDate,expiryDate,status,policyValue,balance,ded,dedInPatient,dedOutPatient,ceiling,ceilingInPatient,ceilingOutPatient
                        }
                    }
                }
            }
            ''',
            headers={"HTTP_AUTHORIZATION": f"Bearer {self.admin_token}"},
            variables={'chfid': self.insuree.chf_id, 'targetDate':self.policy.effective_date.strftime("%Y-%m-%d")}
        )

        content = json.loads(response.content)

        # This validates the status code and if you get errors
        self.assertResponseNoErrors(response)
        self.assertEqual(content['data']['policiesByInsuree']['totalCount'], 1)
        self.assertEqual(UUID(content['data']['policiesByInsuree']['edges'][0]['node']['policyUuid']), UUID(self.policy.uuid))
        
    def test_family_query_with_variables(self):
        response = self.query(
            '''
            query policiesByFamily($familyUuid: String!, $targetDate:  Date! ) {
                policiesByFamily(orderBy: "expiryDate",activeOrLastExpiredOnly: true,familyUuid:$familyUuid ,targetDate: $targetDate,first: 5)
                {
                    totalCount
                    pageInfo { hasNextPage, hasPreviousPage, startCursor, endCursor}
                    edges
                    {
                        node
                        {
                            policyUuid,productCode,productName,officerCode,officerName,enrollDate,effectiveDate,startDate,expiryDate,status,policyValue,balance,ded,dedInPatient,dedOutPatient,ceiling,ceilingInPatient,ceilingOutPatient
                        }
                    }   
                }
            } 
            ''',
            headers={"HTTP_AUTHORIZATION": f"Bearer {self.admin_token}"},
            variables={'familyUuid': str(self.insuree.family.uuid), 'targetDate':self.policy.effective_date.strftime("%Y-%m-%d")}
        )

        content = json.loads(response.content)
        
        

        # This validates the status code and if you get errors
        self.assertResponseNoErrors(response)
        self.assertEqual(content['data']['policiesByFamily']['totalCount'], 1)
        self.assertEqual(UUID(content['data']['policiesByFamily']['edges'][0]['node']['policyUuid']), UUID(self.policy.uuid))

       
   
    def test_insuree_policy_query(self):
        
        
        
        
        
        response = self.query(
            f'''
{{
  policyServiceEligibilityByInsuree(chfId:"{self.insuree.chf_id}", serviceCode:"{self.service.code}")
  {{
    minDateService, serviceLeft, isServiceOk
  }}
}}
            ''',
            headers={"HTTP_AUTHORIZATION": f"Bearer {self.admin_token}"},
        )

        content = json.loads(response.content)

        # This validates the status code and if you get errors
        self.assertResponseNoErrors(response)


        # Add some more asserts if you like
        response = self.query(
            f'''
{{
  policyItemEligibilityByInsuree(chfId:"{self.insuree.chf_id}",itemCode:"{self.item.code}")
  {{
     minDateItem,itemLeft,isItemOk
  }}
}}
            ''',
            headers={"HTTP_AUTHORIZATION": f"Bearer {self.admin_token}"},
        )

        content = json.loads(response.content)

        # This validates the status code and if you get errors
        self.assertResponseNoErrors(response)

        # Add some more asserts if you like
        ...

    def test_mutation_simple(self):
        muuid = "203327cd-501e-41e1-a026-ed742e360081"
        response = self.query(
            f'''
    mutation {{
      createPolicy(
        input: {{
          clientMutationId: "{muuid}"
          clientMutationLabel: "Création de la police ttttt eeeee (123123123) - 2024-06-01 : 2025-05-31"
          
          enrollDate: "2024-04-07"
            startDate: "2024-06-01"
            expiryDate: "2025-05-31"
            value: "10000.00"
            productId: {self.product.id}
            familyId: {self.insuree.family.id}
            officerId: 1
                    }}
        ) {{
            clientMutationId
            internalId
        }}
    }}
            ''',
            headers={"HTTP_AUTHORIZATION": f"Bearer {self.admin_token}"},
            variables={'chfid': self.insuree.chf_id, 'activeOrLastExpiredOnly':True}
        )
        content = self.get_mutation_result(muuid,self.admin_token)
        



import uuid

from core import fields
from core import models as core_models
from core.utils import filter_validity
from core.models import Officer

from django.conf import settings
from django.db import models
from graphql import ResolveInfo
from insuree.models import Family
from product.models import Product


class Policy(core_models.VersionedModel):
    id = models.AutoField(db_column='PolicyID', primary_key=True)
    uuid = models.CharField(db_column='PolicyUUID', max_length=36, default=uuid.uuid4, unique=True)

    stage = models.CharField(db_column='PolicyStage', max_length=1, blank=True, null=True)
    status = models.SmallIntegerField(db_column='PolicyStatus', blank=True, null=True)
    value = models.DecimalField(db_column='PolicyValue', max_digits=18, decimal_places=2, blank=True, null=True)

    family = models.ForeignKey(Family, models.DO_NOTHING, db_column='FamilyID', related_name="policies")
    enroll_date = fields.DateField(db_column='EnrollDate')
    start_date = fields.DateField(db_column='StartDate')
    effective_date = fields.DateField(db_column='EffectiveDate', blank=True, null=True)
    expiry_date = fields.DateField(db_column='ExpiryDate', blank=True, null=True)

    product = models.ForeignKey(Product, models.DO_NOTHING, db_column='ProdID', related_name="policies")
    officer = models.ForeignKey(Officer, models.DO_NOTHING, db_column='OfficerID', blank=True, null=True,
                                related_name="policies")

    offline = models.BooleanField(db_column='isOffline', blank=True, null=True)
    audit_user_id = models.IntegerField(db_column='AuditUserID')
    # row_id = models.BinaryField(db_column='RowID', blank=True, null=True)

    @staticmethod
    def get_query_sum_premium(photo=False):
        return models.Sum(
            'premiums__amount', 
            filter=models.Q(*filter_validity(prefix='premiums__'),premiums__is_photo_fee=photo)
            )
    
    def sum_premiums(self, photo = False):
        return Policy.objects.filter(id=self.id).aggregate(
                sum_premiums=Policy.get_query_sum_premium(photo)
                )['sum_premiums'] or 0

    def claim_ded_rems(self):
        return self.claim_ded_rems

    def is_new(self):
        return not self.stage or self.stage == Policy.STAGE_NEW

    def can_add_insuree(self):
        return self.family.members.filter(validity_to__isnull=True).count() < self.product.max_members
    class Meta:
        managed = True
        db_table = 'tblPolicy'

    STATUS_IDLE = 1
    STATUS_ACTIVE = 2
    STATUS_SUSPENDED = 4
    STATUS_EXPIRED = 8
    STATUS_READY = 16

    STAGE_NEW = 'N'
    STAGE_RENEWED = 'R'

    @classmethod
    def get_queryset(cls, queryset, user):
        queryset = Policy.filter_queryset(queryset)
        # GraphQL calls with an info object while Rest calls with the user itself
        if isinstance(user, ResolveInfo):
            user = user.context.user
        if settings.ROW_SECURITY and user.is_anonymous:
            return queryset.filter(id=-1)
        # TODO: check the access to the policy information but how ?
        #   Policy -> Product -> Location ? Policy -> Insurees -> HF -> Location ?
        # if settings.ROW_SECURITY:
        #     dist = UserDistrict.get_user_districts(user._u)
        #     return queryset.filter(
        #         health_facility__location_id__in=[l.location.id for l in dist]
        #     )
        
        return queryset


class PolicyRenewal(core_models.VersionedModel):
    id = models.AutoField(db_column='RenewalID', primary_key=True)
    uuid = models.CharField(db_column='RenewalUUID', max_length=36, default=uuid.uuid4, unique=True)

    new_officer = models.ForeignKey(Officer, models.DO_NOTHING, db_column='NewOfficerID',
                                    blank=True, null=True, related_name="policy_renewals")
    insuree = models.ForeignKey('insuree.Insuree', models.DO_NOTHING, db_column='InsureeID',
                                related_name='policy_renewals')
    policy = models.ForeignKey('policy.Policy', models.DO_NOTHING, db_column='PolicyID',
                               related_name='policy_renewals')
    new_product = models.ForeignKey(Product, db_column='NewProdID', on_delete=models.DO_NOTHING,
                                       related_name="policy_renewals")

    renewal_prompt_date = fields.DateField(db_column='RenewalPromptDate')
    renewal_date = fields.DateField(db_column='RenewalDate')
    phone_number = models.CharField(db_column='PhoneNumber', max_length=25, blank=True, null=True)
    sms_status = models.SmallIntegerField(db_column='SMSStatus', default=0)  # TODO choices
    renewal_warnings = models.SmallIntegerField(db_column='RenewalWarnings', null=True, blank=True, default=0)  # TODO choices
    response_status = models.IntegerField(db_column='ResponseStatus', null=True, blank=True, default=0)  # TODO Choices
    response_date = fields.DateTimeField(db_column='ResponseDate', null=True, blank=True)

    audit_user_id = models.IntegerField(db_column='AuditCreateUser', null=True, blank=True)

    class Meta:
        managed = True
        db_table = 'tblPolicyRenewals'


class PolicyMutation(core_models.UUIDModel, core_models.ObjectMutation):
    policy = models.ForeignKey(Policy, models.DO_NOTHING,
                                 related_name='mutations')
    mutation = models.ForeignKey(
        core_models.MutationLog, models.DO_NOTHING, related_name='policies')

    class Meta:
        managed = True
        db_table = "policy_PolicyMutation"



from django.utils.translation import gettext as _
from .models import Policy


def validate_idle_policy(policy_input):
    errors = []
    policy_uuid = policy_input.get('uuid')
    if policy_uuid:
        policy = Policy.objects.filter(uuid=policy_uuid, validity_to__isnull=True).first()
        if policy is None:
            return [{
                'message': _("policy.mutation.failed_to_update_policy"),
                'detail': _("policy.validation.id_does_not_exist") % {'id': policy_uuid}
            }]
        errors += check_can_update(policy)
    # TODO: check dates,...
    return errors


def check_can_update(policy):
    if policy.status != Policy.STATUS_IDLE:
        return [{
            'message': _("policy.mutation.failed_to_update_policy"),
            'detail': _("policy.mutation.policy_not_idle")
        }]
    return []



import logging

from policy.services import insert_renewals, update_renewals, policy_renewal_sms

logger = logging.getLogger(__name__)


def get_policies_for_renewal(interval=None, region=None, district=None, ward=None, village=None, officer=None,
                             date_from=None, date_to=None, family_message_template=None, sms_header_template=None):
    """
    Find policies that are due for renewal, add them to the renewal queue, mark the expired policies as expired
    All parameters are optional.
    This method is more a sample than the actual code since it should be heavily customized

    :param interval: number of days before expiration to send renewal
    :param region: region id for which to send the renewals
    :param district: district for which to send the renewals
    :param ward: ward for which to send the renewals
    :param village: village for which to send the renewals
    :param officer: limit renewals to a specific officer
    :param date_from: date range to send renewals
    :param date_to: date range to send renewals
    :param family_message_template: family message template. This a Django template that provides: renewal object,
            district_name, ward_name, village_name, ...
    :param sms_header_template: Also a Django template for the SMS header
    :return: nothing
    """
    for item in [region, district, ward, village]:
        if item:
            location = item
            break
    else:
        location = None
    insert_renewals(date_from, date_to, officer_id=officer, reminding_interval=interval, location_id=location)
    update_renewals()
    sms_queue = policy_renewal_sms(family_message_template, date_from, date_to, sms_header_template)
    for sms in sms_queue:
        send_sms(sms)


def send_sms(sms):
    """
    This method is quite specific to the SMS provider. It would be a good idea to adapt the above task to suit
    the needs of the gateway, its processing status etc.
    :param sms: sms queue item, contains phone, sms_message, index
    """
    logger.warning("Sending an SMS needs a defined gateway, pretending to send to %s:\n%s", sms.phone, sms.sms_message)



from django.utils.translation import gettext as _
from django.db.models import Q,Count
import datetime as py_datetime
from decimal import *
from .models import Policy

from core.apps import CoreConfig
from dateutil.relativedelta import relativedelta
from core.apps import CoreConfig


def cycle_start(product, cycle, ref_date):
    c = getattr(product, "start_cycle_%s" % (cycle + 1), None)
    if not c:
        return None
    if CoreConfig.secondary_calendar == 'Nepal':
        import nepali_datetime
        nepali_start = nepali_datetime.datetime.strptime("%s-%s" % (c, nepali_datetime.date.today().year), '%d-%m-%Y')
        start = nepali_start.to_datetime_date()
    else:
        start = py_datetime.datetime.strptime("%s-%s" % (c, ref_date.year), '%d-%m-%Y')
    if ref_date <= start:
        return start


def set_start_date(policy):
    from core import datetime, datetimedelta
    product = policy.product
    ref_enroll_date = policy.enroll_date
    if policy.stage == Policy.STAGE_NEW and product.administration_period:
        ref_enroll_date = (
                datetime.date.from_ad_date(ref_enroll_date) +
                datetimedelta(months=product.administration_period)
        ).to_ad_date()

    if not product.has_cycle():
        policy.start_date = ref_enroll_date
        return

    grace = 0
    if policy.stage == Policy.STAGE_NEW and product.grace_period_enrolment:
        grace = product.grace_period_enrolment
    elif policy.stage == Policy.STAGE_RENEWED and product.grace_period_renewal:
        grace = product.grace_period_renewal

    ref_date = (datetime.date.from_ad_date(ref_enroll_date) - datetimedelta(months=grace)).to_ad_date()
    for i in range(4):
        start = cycle_start(product, i, ref_date)
        if start:
            policy.start_date = datetime.date.from_ad_date(start)
            return
    policy.start_date = datetime.date.from_ad_date(py_datetime.datetime.strptime(
        "%s-%s" % (product.start_cycle_1, ref_date.year + 1),
        '%d-%m-%Y'
    ))


def set_expiry_date(policy):
    product = policy.product
    from core import datetime, datetimedelta

    insurance_period = datetimedelta(
        months=product.insurance_period) if product.insurance_period % 12 != 0 else datetimedelta(
        years=product.insurance_period // 12)
    policy.expiry_date = (
            datetime.date.from_ad_date(policy.start_date) +
            insurance_period -
            datetimedelta(days=1)
    ).to_ad_date()


def family_counts(product, family):
    adults = 0
    other_adults = 0
    extra_adults = 0
    children = 0
    other_children = 0
    extra_children = 0
    total = 0
    date_threshold = py_datetime.date.today()- relativedelta(years= CoreConfig.age_of_majority)
    counts= family.members.filter(
        Q(validity_to__isnull=True),
    ).aggregate(
        adults=Count('id',filter=Q(Q(dob__lt=date_threshold) & ~Q(relationship_id=7))),
        children=Count('id', filter=Q(Q(dob__gte=date_threshold) & ~Q(relationship_id=7))),
        other_children=Count('id', filter=Q(relationship_id=7, dob__gte=date_threshold)),
        other_adults=Count('id', filter=Q(relationship_id=7, dob__lt=date_threshold))
    )
    adults = counts['adults'] 
    children =  counts['children'] 
    other_children =  counts['other_children'] 
    other_adults =  counts['other_adults'] 
    
    over_children = 0
    over_adults = 0

    if product.max_members:
        over_adults = max(0,adults   - product.max_members)
        over_children = max(0,adults  + children  - over_adults - product.max_members)
        over_other_children = max(0,adults + children + other_children -over_adults -over_children- product.max_members)
        over_other_adults = max(0,adults  + other_adults + children + other_children -over_adults - over_other_children - over_children - product.max_members)
        
    # remove over from count
    children -= over_children
    adults -= over_adults
    other_children -= over_other_children
    other_adults -= over_other_adults
    if product.threshold:   
        extra_adults = max(0, adults -   product.threshold)
        extra_children = max(0,children + adults - extra_adults - product.threshold)


    return {
        "adults":adults -extra_adults, # adult part of the "lump sum"
        "extra_adults": extra_adults, # adult not part of the "lump sum" because of threshold
        "other_adults": other_adults , # adult never of the "lump sum"
        "children": children-extra_children,  # children part of the "lump sum"
        "extra_children": extra_children, # children never part of the "lump sum" because of threshold
        "other_children": other_children,# children never part of the "lump sum"
        "total": adults  + other_adults + children  +  other_children,
    }


def get_attr(product, attr):
    #  getattr(product, attr, 0)... returns None if the attr is there (with None as value!)
    value = getattr(product, attr, 0)
    return value if value else 0


def sum_contributions(product, f_counts):
    contributions = 0
    premium_adult = get_attr(product, 'premium_adult')
    premium_child = get_attr(product, 'premium_child')
    if product.lump_sum:
        contributions = product.lump_sum
        contributions += f_counts["extra_adults"] * premium_adult
        contributions += f_counts["extra_children"] * premium_child
    else:
        contributions += f_counts["adults"] * premium_adult
        contributions += f_counts["children"] * premium_child
    contributions += f_counts["other_adults"] * premium_adult
    contributions += f_counts["other_children"] * premium_child
    return contributions


def sum_general_assemblies(product, f_counts):
    if product.general_assembly_lump_sum:
        return product.general_assembly_lump_sum
    return f_counts["total"] * get_attr(product, 'general_assembly_fee')


def sum_registrations(policy, product, f_counts):
    if policy.stage != Policy.STAGE_NEW:
        return 0
    if product.registration_lump_sum:
        return product.registration_lump_sum
    return f_counts["total"] * get_attr(product, 'registration_fee')


def discount_new(policy):
    product = policy.product
    if product.has_enrolment_discount() and product.has_cycle():
        from core import datetime, datetimedelta
        min_discount_date = (
                datetime.date.from_ad_date(policy.start_date) - datetimedelta(months=product.enrolment_discount_period)
        ).to_ad_datetime()
        if policy.enroll_date <= min_discount_date:
            policy.value -= policy.value * product.enrolment_discount_perc / 100


def discount_renew(policy, prev_policy):
    product = policy.product
    if product.has_renewal_discount():
        from core import datetime, datetimedelta
        min_discount_date = (
                datetime.date.from_ad_date(prev_policy.expiry_date) +
                datetimedelta(days=1) -
                datetimedelta(months=product.renewal_discount_period)
        ).to_ad_datetime()
        if policy.enroll_date <= min_discount_date:
            policy.value -= policy.value * product.renewal_discount_perc / 100


def discount(policy, prev_policy):
    if policy.stage == Policy.STAGE_NEW:
        discount_new(policy)
    elif policy.stage == Policy.STAGE_RENEWED:
        discount_renew(policy, prev_policy)


def set_value(policy, family, prev_policy):
    product = policy.product
    f_counts = family_counts(policy.product, family)
    contributions = sum_contributions(product, f_counts)
    general_assembly = sum_general_assemblies(product, f_counts)
    registration = sum_registrations(policy, product, f_counts)
    policy.value = Decimal(contributions + general_assembly + registration)
    discount(policy, prev_policy)


def policy_values(policy, family, prev_policy):
    members = family.members.filter(validity_to__isnull=True).count()
    max_members = policy.product.max_members
    above_max = max(0, members - max_members)
    warnings = []
    if above_max:
        warnings.append(_("policy.validation.members_count_above_max") % {'max': max_members, 'count': members})
    set_start_date(policy)
    set_expiry_date(policy)
    set_value(policy, family, prev_policy)
    return policy, warnings



import logging
from dataclasses import dataclass
from datetime import datetime as py_datetime, date as py_date

import core
from claim.models import ClaimService, Claim, ClaimItem
from django import dispatch
from django.core.exceptions import PermissionDenied, ValidationError
from django.db import connection
from django.db.models import Q, Count, Min, Max, Sum, F
from django.db.models.functions import Coalesce
from django.template import Template, Context
from django.utils.translation import gettext as _
from graphene.utils.str_converters import to_snake_case

from policy.utils import get_queryset_valid_at_date
from core.signals import register_service_signal
from insuree.models import Insuree, Family, InsureePolicy
from insuree.services import create_insuree_renewal_detail
from medical.models import Service, Item
from policy.apps import PolicyConfig
from policy.utils import MonthsAdd

from .models import Policy, PolicyRenewal

logger = logging.getLogger(__name__)


def reset_policy_before_update(policy):
    policy.enroll_date = None
    policy.start_date = None
    policy.expiry_date = None
    policy.value = None
    policy.product_id = None
    policy.family_id = None
    policy.officer_id = None


class PolicyService:
    def __init__(self, user):
        self.user = user

    @register_service_signal('policy_service.create_or_update')
    def update_or_create(self, data, user): 
        if isinstance(data['enroll_date'], str):
            data['enroll_date'] = py_datetime.strptime(data['enroll_date'], "%Y-%m-%d").date()
        policy_uuid = data.get('uuid', None)
        if 'enroll_date' in data and data['enroll_date'] > py_date.today():
            raise ValidationError("policy.enroll_date_in_the_future")
        if policy_uuid:
            return self.update_policy(data, user)
        else:
            return self.create_policy(data, user)

    @register_service_signal('policy_service.update')
    def update_policy(self, data, user):
        if "is_paid" in data:
            data.pop("is_paid")
        data = self._clean_mutation_info(data)
        policy_uuid = data.pop('uuid') if 'uuid' in data else None
        policy = Policy.objects.get(uuid=policy_uuid)
        policy.save_history()
        reset_policy_before_update(policy)
        [setattr(policy, key, data[key]) for key in data]
        policy.save()
        update_insuree_policies(policy, user.id_for_audit)
        return policy

    @register_service_signal('policy_service.create')
    def create_policy(self, data, user):
        is_paid = data.pop("is_paid", False)
        receipt = data.pop("receipt", None)
        payer_uuid = data.pop("payer_uuid", None)
        data = self._clean_mutation_info(data)
        policy = Policy.objects.create(**data)
        if receipt is not None:
            from contribution.services import check_unique_premium_receipt_code_within_product
            is_invalid = check_unique_premium_receipt_code_within_product(code=receipt, policy_uuid=policy.uuid)
            if is_invalid:
                raise ValidationError("Receipt already exist for a given product.")
        else:
            receipt = self.generate_contribution_receipt(policy.product, policy.enroll_date)
        policy.save()
        update_insuree_policies(policy, user.id_for_audit)
        if is_paid:
            from contribution.gql_mutations import premium_action
            premium_data = {"policy_uuid": policy.uuid, "amount": policy.value,
                            "receipt": receipt, "pay_date": data["enroll_date"], "pay_type": "C"}
            if payer_uuid is not None:
                premium_data["payer_uuid"] = payer_uuid
            premium_action(premium_data, user)
        return policy

    def generate_contribution_receipt(self, product, enroll_date):
        from contribution.models import Premium
        code_length = PolicyConfig.contribution_receipt_length
        if not code_length and type(code_length) is not int:
            raise ValueError("Invalid config for `generate_contribution_receipt`, expected `code_length` value.")
        prefix = "RE-" + str(product.code) + "-" + str(enroll_date) + "-"
        last_contribution = Premium.objects.filter(validity_to__isnull=True, receipt__icontains=prefix)
        code = 0
        if last_contribution:
            code = int(last_contribution.latest('receipt').receipt[-code_length:])
        return prefix + str(code + 1).zfill(code_length)

    def _clean_mutation_info(self, data):
        if "client_mutation_id" in data:
            data.pop('client_mutation_id')
        if "client_mutation_label" in data:
            data.pop('client_mutation_label')
        return data

    def set_suspended(self, user, policy):
        try:
            policy.save_history()
            policy.status = Policy.STATUS_SUSPENDED
            policy.audit_user_id = user.id_for_audit
            policy.save()
            return []
        except Exception as exc:
            return {
                'title': policy.uuid,
                'list': [{
                    'message': _("policy.mutation.failed_to_suspend_policy") % {'uuid': policy.uuid},
                    'detail': policy.uuid}]
            }

    def set_deleted(self, policy):
        try:
            insuree_policies = InsureePolicy.objects.filter(policy=policy)
            for insuree_policy in insuree_policies:
                insuree_policy.delete_history()
            policy.delete_history()
            return []
        except Exception as exc:
            return {
                'title': policy.uuid,
                'list': [{
                    'message': _("policy.mutation.failed_to_change_status_of_policy") % {'policy': str(policy)},
                    'detail': policy.uuid}]
            }


@core.comparable
class ByInsureeRequest(object):

    def __init__(self, chf_id, active_or_last_expired_only=False, show_history=False, order_by=None, target_date=None):
        self.chf_id = chf_id
        self.active_or_last_expired_only = active_or_last_expired_only
        self.show_history = show_history
        self.order_by = order_by
        self.target_date = target_date

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.__dict__ == other.__dict__


@core.comparable
class ByFamilyOrInsureeResponseItem(object):

    def __init__(self,
                 policy_id,
                 policy_uuid,
                 policy_value,
                 product_code,
                 product_name,
                 start_date,
                 enroll_date,
                 effective_date,
                 expiry_date,
                 officer_code,
                 officer_name,
                 status,
                 ded,
                 ded_in_patient,
                 ded_out_patient,
                 ceiling,
                 ceiling_in_patient,
                 ceiling_out_patient,
                 balance,
                 validity_from,
                 validity_to,
                 max_installments
                 ):
        self.policy_id = policy_id
        self.policy_uuid = policy_uuid
        self.policy_value = policy_value
        self.product_code = product_code
        self.product_name = product_name
        self.start_date = start_date
        self.enroll_date = enroll_date
        self.effective_date = effective_date
        self.expiry_date = expiry_date
        self.officer_code = officer_code
        self.officer_name = officer_name
        self.status = status
        self.ded = ded
        self.ded_in_patient = ded_in_patient
        self.ded_out_patient = ded_out_patient
        self.ceiling = ceiling
        self.ceiling_in_patient = ceiling_in_patient
        self.ceiling_out_patient = ceiling_out_patient
        self.balance = balance
        self.validity_from = validity_from
        self.validity_to = validity_to
        self.max_installments = max_installments

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.__dict__ == other.__dict__


@core.comparable
class ByInsureeResponse(object):

    def __init__(self, by_insuree_request, items):
        self.by_insuree_request = by_insuree_request
        self.items = items

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.__dict__ == other.__dict__


class FilteredPoliciesService(object):
    def __init__(self, user):
        self.user = user

    @staticmethod
    def _to_item(row):
        ceiling = None
        ceiling_ip = None
        ceiling_op = None
        if row.product.max_treatment:
            ceiling = row.product.max_treatment
        if row.product.max_ip_treatment:
            ceiling_ip = row.product.max_ip_treatment
        if row.product.max_op_treatment:
            ceiling_op = row.product.max_ip_treatment
        if row.product.max_insuree:
            ceiling = row.product.max_insuree - (row.total_rem_g if row.total_rem_g else 0)
        else:
            if row.product.max_ip_insuree:
                ceiling_ip = row.product.max_ip_insuree - (row.total_rem_ip if row.total_rem_ip else 0)
            if row.product.max_op_insuree:
                ceiling_op = row.product.max_op_insuree - (row.total_rem_op if row.total_rem_op else 0)

        members_count = row.family.members.count()
        threshold = row.product.threshold if row.product.threshold else 0
        total_rem_g = row.total_rem_g if row.total_rem_g else 0
        total_rem_ip = row.total_rem_ip if row.total_rem_ip else 0
        total_rem_op = row.total_rem_op if row.total_rem_op else 0
        extra_member = row.product.max_policy_extra_member if row.product.max_policy_extra_member else 0
        extra_member_ip = row.product.max_policy_extra_member_ip if row.product.max_policy_extra_member_ip else 0
        extra_member_op = row.product.max_policy_extra_member_op if row.product.max_policy_extra_member_op else 0


        if row.product.max_policy:
            max_policy = row.product.max_policy
            if members_count > threshold:
                max_policy += (members_count - threshold) * extra_member
            ceiling = max_policy - total_rem_g
        else:
            ceiling_ip = 0
            if row.product.max_ip_policy:
                max_ip_policy = row.product.max_ip_policy
                if members_count > threshold:
                    max_ip_policy += (members_count - threshold) * extra_member_ip
                ceiling_ip = max_ip_policy - total_rem_ip

            ceiling_op = 0
            if row.product.max_op_policy:
                max_op_policy = row.product.max_op_policy
                if members_count > threshold:
                    max_op_policy += (members_count - threshold) * extra_member_op
                ceiling_op = max_op_policy - total_rem_op

        balance = row.value
        if row.total_ded_g:
            balance -= row.total_ded_g

        return ByFamilyOrInsureeResponseItem(
            policy_id=row.id,
            policy_uuid=row.uuid,
            policy_value=row.value,
            product_code=row.product.code,
            product_name=row.product.name,
            start_date=row.start_date,
            enroll_date=row.enroll_date,
            effective_date=row.effective_date,
            expiry_date=row.expiry_date,
            officer_code=row.officer.code if row.officer else None,
            officer_name=row.officer.name() if row.officer else None,
            status=row.status,
            ded=row.total_ded_g,
            ded_in_patient=row.total_ded_ip,
            ded_out_patient=row.total_ded_op,
            ceiling=ceiling,
            ceiling_in_patient=ceiling_ip,
            ceiling_out_patient=ceiling_op,
            balance=balance,
            validity_from=row.validity_from,
            validity_to=row.validity_to,
            max_installments=row.product.max_installments,
        )

    def build_query(self, req):
        # TODO: prevent direct dependency on claim_ded structure?
        res = Policy.objects\
            .prefetch_related('product') \
            .prefetch_related('officer') \
            .annotate(total_ded_g=Sum('claim_ded_rems__ded_g')) \
            .annotate(total_ded_ip=Sum('claim_ded_rems__ded_ip')) \
            .annotate(total_ded_op=Sum('claim_ded_rems__ded_op')) \
            .annotate(total_rem_g=Sum('claim_ded_rems__rem_g')) \
            .annotate(total_rem_op=Sum('claim_ded_rems__rem_op')) \
            .annotate(total_rem_ip=Sum('claim_ded_rems__rem_ip')) \
            .annotate(total_rem_consult=Sum('claim_ded_rems__rem_consult')) \
            .annotate(total_rem_surgery=Sum('claim_ded_rems__rem_surgery')) \
            .annotate(total_rem_delivery=Sum('claim_ded_rems__rem_delivery')) \
            .annotate(total_rem_hospitalization=Sum('claim_ded_rems__rem_hospitalization')) \
            .annotate(total_rem_antenatal=Sum('claim_ded_rems__rem_antenatal'))
        res.query.group_by = ['id']
        if hasattr(req, 'chf_id'):
            res= res.filter(insuree_policies__insuree__chf_id = req.chf_id)
        if not req.show_history:
            if req.target_date: 
                res = res.filter(*core.filter_validity(), expiry_date__gt = req.target_date, effective_date__lte = req.target_date)
            else:
                res = res.filter(*core.filter_validity())
        if req.active_or_last_expired_only:
            # sort on status, so that any active policy (status = 2) pops up...
            res = res.annotate(not_null_expiry_date=Coalesce('expiry_date', py_date.max)) \
                .annotate(not_null_validity_to=Coalesce('validity_to', py_datetime.max)) \
                .order_by('product__code', 'status', '-not_null_expiry_date', '-not_null_validity_to', '-validity_from')
        return res


class ByInsureeService(FilteredPoliciesService):
    def __init__(self, user):
        super(ByInsureeService, self).__init__(user)

    def request(self, by_insuree_request):
        res = self.build_query(by_insuree_request)
        res = res.filter(insuree_policies__insuree__chf_id=by_insuree_request.chf_id)
        if by_insuree_request.active_or_last_expired_only:
            products = {}
            for policy in res:
                if policy.status == Policy.STATUS_IDLE or policy.status == Policy.STATUS_READY:
                    products['policy.product.code-%s' % policy.uuid] = policy
                elif policy.product.code not in products.keys():
                    products[policy.product.code] = policy
            res = products.values()
        items = [FilteredPoliciesService._to_item(x) for x in res]
        # possible improvement: sort via the ORM
        # ... but beware of the active_or_last_expired_only filtering!
        order_attr = to_snake_case(by_insuree_request.order_by if by_insuree_request.order_by else "expiry_date")
        desc = False
        if order_attr.startswith('-'):
            order_attr = order_attr[1:]
            desc = True
        items = sorted(items, key=lambda x: getattr(x, order_attr), reverse=desc)
        return ByInsureeResponse(
            by_insuree_request=by_insuree_request,
            items=items
        )


@core.comparable
class ByFamilyRequest(object):

    def __init__(self, family_uuid, active_or_last_expired_only=False, show_history=False, order_by=None, target_date=None):
        self.family_uuid = family_uuid
        self.active_or_last_expired_only = active_or_last_expired_only
        self.show_history = show_history
        self.order_by = order_by
        self.target_date = target_date

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.__dict__ == other.__dict__


@core.comparable
class ByFamilyResponse(object):

    def __init__(self, by_family_request, items):
        self.by_family_request = by_family_request
        self.items = items

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.__dict__ == other.__dict__


class ByFamilyService(FilteredPoliciesService):
    def __init__(self, user):
        super(ByFamilyService, self).__init__(user)

    def request(self, by_family_request):
        res = self.build_query(by_family_request)
        res = res.filter(family__uuid=by_family_request.family_uuid)
        # .distinct('product__code') >> DISTINCT ON fields not supported by MS-SQL
        if by_family_request.active_or_last_expired_only:
            products = {}
            for policy in res:
                if policy.status == Policy.STATUS_IDLE or policy.status == Policy.STATUS_READY:
                    products['policy.product.code-%s' % policy.uuid] = policy
                elif policy.product.code not in products.keys():
                    products[policy.product.code] = policy
            res = products.values()
        items = tuple(
            map(lambda x: FilteredPoliciesService._to_item(x), res)
        )
        return ByFamilyResponse(
            by_family_request=by_family_request,
            items=items
        )


# --- ELIGIBILITY --
# TODO: should become "BY FAMILY":
# Eligibility is calculated from a Policy
# ... which is bound to a Family (same remark as ByInsureeService)
# -------------------
@core.comparable
class EligibilityRequest(object):

    def __init__(self, chf_id, service_code=None, item_code=None):
        self.chf_id = chf_id
        self.service_code = service_code
        self.item_code = item_code

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.__dict__ == other.__dict__


class EligibilityResponse(object):

    def __init__(self, eligibility_request, prod_id=None, total_admissions_left=0, total_visits_left=0,
                 total_consultations_left=0, total_surgeries_left=0,
                 total_deliveries_left=0, total_antenatal_left=0, consultation_amount_left=0, surgery_amount_left=0,
                 delivery_amount_left=0,
                 hospitalization_amount_left=0, antenatal_amount_left=0,
                 min_date_service=None, min_date_item=None, service_left=0, item_left=0, is_item_ok=False,
                 is_service_ok=False, final=False):
        self.eligibility_request = eligibility_request
        self.prod_id = prod_id
        self.total_admissions_left = total_admissions_left
        self.total_visits_left = total_visits_left
        self.total_consultations_left = total_consultations_left
        self.total_surgeries_left = total_surgeries_left
        self.total_deliveries_left = total_deliveries_left
        self.total_antenatal_left = total_antenatal_left
        self.consultation_amount_left = consultation_amount_left
        self.surgery_amount_left = surgery_amount_left
        self.delivery_amount_left = delivery_amount_left
        self.hospitalization_amount_left = hospitalization_amount_left
        self.antenatal_amount_left = antenatal_amount_left
        self.min_date_service = min_date_service
        self.min_date_item = min_date_item
        self.service_left = service_left
        self.item_left = item_left
        self.is_item_ok = is_item_ok
        self.is_service_ok = is_service_ok
        self.final = final

    def __eq__(self, other):
        if not isinstance(other, self.__class__):
            return False

        def str_none(x):
            str(x) if x else x

        # Comparison should take into account the date vs AdDate
        return (
                self.eligibility_request == other.eligibility_request and
                self.prod_id == other.prod_id and
                self.total_admissions_left == other.total_admissions_left and
                self.total_visits_left == other.total_visits_left and
                self.total_consultations_left == other.total_consultations_left and
                self.total_surgeries_left == other.total_surgeries_left and
                self.total_deliveries_left == other.total_deliveries_left and
                self.total_antenatal_left == other.total_antenatal_left and
                self.consultation_amount_left == other.consultation_amount_left and
                self.surgery_amount_left == other.surgery_amount_left and
                self.delivery_amount_left == other.delivery_amount_left and
                self.hospitalization_amount_left == other.hospitalization_amount_left and
                self.antenatal_amount_left == other.antenatal_amount_left and
                str_none(self.min_date_service) == str_none(other.min_date_service) and
                str_none(self.min_date_item) == str_none(other.min_date_item) and
                self.service_left == other.service_left and
                self.item_left == other.item_left and
                self.is_item_ok == other.is_item_ok and
                self.is_service_ok == other.is_service_ok)

    def __str__(self):
        return f"Eligibility for {self.eligibility_request} gave product {self.prod_id} " \
               f"with item/svc ok {self.is_item_ok}/{self.is_service_ok} " \
               f" left: {self.item_left}/{self.service_left}"

    def __repr__(self):
        return self.__str__()


signal_eligibility_service_before = dispatch.Signal(["user", "request", "response"])
signal_eligibility_service_after = dispatch.Signal(["user", "request", "response"])


class EligibilityService(object):
    def __init__(self, user):
        self.user = user
        self.service = NativeEligibilityService(user)

    def request(self, request):
        if not self.user or not self.user.has_perms(PolicyConfig.gql_query_eligibilities_perms):
            raise PermissionDenied()

        # The response is passed along in signals and functions. Setting the final parameter will stop
        response = EligibilityResponse(eligibility_request=request)
        responses = signal_eligibility_service_before.send(
            self.__class__, user=self.user, request=request, response=response)
        response = EligibilityService._get_final_response(responses, "before", response)

        if not response.final and not PolicyConfig.default_eligibility_disabled:
            response = self.service.request(request, response)

        if not response.final:
            responses = signal_eligibility_service_after.send(
                self.__class__, user=self.user, request=request, response=response)
            response = EligibilityService._get_final_response(responses, "after", response)

        return response

    @classmethod
    def _get_final_response(cls, responses, sig_name, default_response):
        if responses is None or len(responses) == 0:
            return default_response
        final_responses = [r for f, r in responses if r.final]
        if len(final_responses) > 0:
            if len(final_responses) > 1:
                logger.warning("Eligibility service got more than one final *%s* signal response: %s",
                               sig_name,
                               [f for f, r in responses if r.final])
            return final_responses[0]
        else:
            return responses[-1]


class StoredProcEligibilityService(object):
    def __init__(self, user):
        self.user = user

    def request(self, req, response):
        with connection.cursor() as cur:
            sql = """\
                DECLARE @MinDateService DATE, @MinDateItem DATE,
                        @ServiceLeft INT, @ItemLeft INT,
                        @isItemOK BIT, @isServiceOK BIT;
                EXEC [dbo].[uspServiceItemEnquiry] @CHFID = %s, @ServiceCode = %s, @ItemCode = %s,
                     @MinDateService = @MinDateService OUTPUT, @MinDateItem = @MinDateItem OUTPUT,
                     @ServiceLeft = @ServiceLeft OUTPUT, @ItemLeft = @ItemLeft OUTPUT,
                     @isItemOK = @isItemOK OUTPUT, @isServiceOK = @isServiceOK OUTPUT;
                SELECT @MinDateService, @MinDateItem, @ServiceLeft, @ItemLeft, @isItemOK, @isServiceOK
            """
            cur.execute(sql, (req.chf_id,
                              req.service_code,
                              req.item_code))
            res = cur.fetchone()  # retrieve the stored proc @Result table
            if res is None:
                return response

            (prod_id, total_admissions_left, total_visits_left, total_consultations_left, total_surgeries_left,
             total_deliveries_left, total_antenatal_left, consultation_amount_left, surgery_amount_left,
             delivery_amount_left,
             hospitalization_amount_left, antenatal_amount_left) = res
            cur.nextset()
            (min_date_service, min_date_item, service_left,
             item_left, is_item_ok, is_service_ok) = cur.fetchone()
            return EligibilityResponse(
                eligibility_request=req,
                prod_id=prod_id or None,
                total_admissions_left=total_admissions_left,
                total_visits_left=total_visits_left,
                total_consultations_left=total_consultations_left,
                total_surgeries_left=total_surgeries_left,
                total_deliveries_left=total_deliveries_left,
                total_antenatal_left=total_antenatal_left,
                consultation_amount_left=consultation_amount_left,
                surgery_amount_left=surgery_amount_left,
                delivery_amount_left=delivery_amount_left,
                hospitalization_amount_left=hospitalization_amount_left,
                antenatal_amount_left=antenatal_amount_left,
                min_date_service=min_date_service,
                min_date_item=min_date_item,
                service_left=service_left,
                item_left=item_left,
                is_item_ok=is_item_ok is True,
                is_service_ok=is_service_ok is True
            )


class NativeEligibilityService(object):
    def __init__(self, user):
        self.user = user


    def get_eligibility(self, insuree, item_or_service, model, req, now):
        if insuree.is_adult():
            waiting_period_field = f"policy__product__{item_or_service}s__waiting_period_adult"
            limit_field = f"policy__product__{item_or_service}s__limit_no_adult"
        else:
            waiting_period_field = f"policy__product__{item_or_service}s__waiting_period_child"
            limit_field = f"policy__product__{item_or_service}s__limit_no_child"

        item_or_service_code = req.service_code
        if item_or_service == "item":
            item_or_service_code = req.item_code

        # TODO validity is checked but should be optional in get_queryset
        item_or_service_obj = model.get_queryset(None, self.user).get(code__iexact=item_or_service_code)

        # Beware that MonthAdd() is in Gregorian calendar, not Nepalese or anything else
        queryset_item_or_service = InsureePolicy.objects\
            .filter(validity_to__isnull=True)\
            .filter(policy__validity_to__isnull=True)\
            .filter(**{f"policy__product__{item_or_service}s__validity_to__isnull": True},
                    **{f"policy__product__{item_or_service}s__{item_or_service}_id": item_or_service_obj.id}) \
            .filter(policy__status=Policy.STATUS_ACTIVE) \
            .filter(insuree=insuree) \
            .filter(Q(insuree__claim__validity_to__isnull=True,
                      **{f"insuree__claim__{item_or_service}s__{item_or_service}_id": item_or_service_obj.id})
                    & Q(**{f"insuree__claim__{item_or_service}s__validity_to__isnull": True})
                    & (Q(**{f"insuree__claim__{item_or_service}s__status": ClaimItem.STATUS_PASSED})
                       | Q(**{f"insuree__claim__{item_or_service}s__status__isnull": True}))
                    & (Q(insuree__claim__status__gt=Claim.STATUS_ENTERED)
                       | Q(insuree__claim__status__isnull=True))) \
            .values("effective_date",
                    "policy__product_id",
                    waiting_period=F(waiting_period_field),
                    limit_no=F(limit_field)) \
            .annotate(min_date=MonthsAdd("effective_date", Coalesce(F(waiting_period_field), 0))) \
            .annotate(count=Sum(Coalesce(
                                f"insuree__claim__{item_or_service}s__qty_approved",
                                f'insuree__claim__{item_or_service}s__qty_provided'
                            ))) \
            .annotate(left=F("limit_no") - F("count"))

        min_date_qs = queryset_item_or_service.aggregate(
            min_date_lte=Min("min_date", filter=Q(min_date__lte=now)),
            min_date_all=Min("min_date"),
        )
        from core import datetime
        min_date_item = datetime.date.from_ad_date(min_date_qs["min_date_lte"]
                                                   if min_date_qs["min_date_lte"]
                                                   else min_date_qs["min_date_all"])

        if queryset_item_or_service.filter(min_date__lte=now).filter(left__isnull=True).order_by('-validity_from').first():
            items_or_services_left = None
        else:
            items_or_services_left = queryset_item_or_service\
                .filter(Q(min_date__isnull=True) | Q(min_date__lte=now))\
                .aggregate(Max("left"))["left__max"]

        return item_or_service_obj, min_date_item, items_or_services_left


    def request(self, req, response):
        insuree = Insuree.get_queryset(None, self.user)\
            .filter(validity_to__isnull=True)\
            .get(chf_id=req.chf_id)  # Will throw an exception if not found
        now = core.datetime.datetime.now()
        eligibility = response

        if req.service_code:
            service, min_date_service, services_left = self.get_eligibility(insuree, "service", Service, req, now)
        else:
            service = None
            services_left = None
            min_date_service = None
        eligibility.min_date_service = min_date_service
        eligibility.service_left = services_left

        if req.item_code:
            item, min_date_item, items_left = self.get_eligibility(insuree, "item", Item, req, now)
        else:
            item = None
            items_left = None
            min_date_item = None
        eligibility.min_date_item = min_date_item
        eligibility.item_left = items_left

        def get_total_filter(category):
            return (
                Q(insuree__claim__category=category)
                & Q(insuree__validity_to__isnull=True)  # Not sure this one is necessary
                & Q(insuree__claim__validity_to__isnull=True)
                & Q(insuree__claim__services__validity_to__isnull=True)
                & Q(insuree__claim__status__gt=Claim.STATUS_ENTERED)
                & (Q(insuree__claim__services__rejection_reason=0)
                   | Q(insuree__claim__services__rejection_reason__isnull=True))
            )

        # InsPol -> Policy -> Product -> dedrem
        result = InsureePolicy.objects \
            .filter(policy__product__validity_to__isnull=True) \
            .filter(policy__validity_to__isnull=True) \
            .filter(validity_to__isnull=True) \
            .filter(insuree=insuree) \
            .values("policy__product_id",
                    "policy__product__max_no_surgery",
                    "policy__product__max_amount_surgery",
                    "policy__product__max_amount_consultation",
                    "policy__product__max_amount_surgery",
                    "policy__product__max_amount_delivery",
                    "policy__product__max_amount_antenatal",
                    "policy__product__max_amount_hospitalization",
                    ) \
            .annotate(total_admissions=Coalesce(Count("insuree__claim",
                                                      filter=get_total_filter(Service.CATEGORY_HOSPITALIZATION),
                                                      distinct=True), 0)) \
            .annotate(total_admissions_left=F("policy__product__max_no_hospitalization")
                                            - F("total_admissions")) \
            .annotate(total_consultations=Coalesce(Count("insuree__claim",
                                                         filter=get_total_filter(Service.CATEGORY_CONSULTATION),
                                                         distinct=True), 0)) \
            .annotate(total_consultations_left=F("policy__product__max_no_consultation")
                                               - F("total_consultations")) \
            .annotate(total_surgeries=Coalesce(Count("insuree__claim",
                                                     filter=get_total_filter(Service.CATEGORY_SURGERY),
                                                     distinct=True), 0)) \
            .annotate(total_surgeries_left=F("policy__product__max_no_surgery") - F("total_surgeries")) \
            .annotate(total_deliveries=Coalesce(Count("insuree__claim",
                                                      filter=get_total_filter(Service.CATEGORY_DELIVERY),
                                                      distinct=True), 0)) \
            .annotate(total_deliveries_left=F("policy__product__max_no_delivery") - F("total_deliveries")) \
            .annotate(total_antenatal=Coalesce(Count("insuree__claim",
                                                     filter=get_total_filter(Service.CATEGORY_ANTENATAL),
                                                     distinct=True), 0)) \
            .annotate(total_antenatal_left=F("policy__product__max_no_antenatal") - F("total_antenatal")) \
            .annotate(total_visits=Coalesce(Count("insuree__claim",
                                                  filter=get_total_filter(Service.CATEGORY_VISIT),
                                                  distinct=True), 0)) \
            .annotate(total_visits_left=F("policy__product__max_no_visits") - F("total_visits")) \
            .order_by('-expiry_date')\
            .first()

        if result is None:
            eligibility.total_admissions_left = 0
            eligibility.total_consultations_left = 0
            eligibility.total_surgeries_left = 0
            eligibility.total_deliveries_left = 0
            eligibility.total_antenatal_left = 0
            eligibility.total_visits_left = 0
            eligibility.surgery_amount_left = 0
            eligibility.consultation_amount_left = 0
            eligibility.delivery_amount_left = 0
            eligibility.antenatal_amount_left = 0
            eligibility.hospitalization_amount_left = 0
            eligibility.is_item_ok = False
            eligibility.is_service_ok = False
            return eligibility

        eligibility.prod_id = result["policy__product_id"]
        total_admissions_left = result["total_admissions_left"] \
            if result["total_admissions_left"] is None or result["total_admissions_left"] >= 0 else 0
        total_consultations_left = result["total_consultations_left"] \
            if result["total_consultations_left"] is None or result["total_consultations_left"] >= 0 else 0
        total_surgeries_left = result["total_surgeries_left"] \
            if result["total_surgeries_left"] is None or result["total_surgeries_left"] >= 0 else 0
        total_deliveries_left = result["total_deliveries_left"] \
            if result["total_deliveries_left"] is None or result["total_deliveries_left"] >= 0 else 0
        total_antenatal_left = result["total_antenatal_left"] \
            if result["total_antenatal_left"] is None or result["total_antenatal_left"] >= 0 else 0
        total_visits_left = result["total_visits_left"] \
            if result["total_visits_left"] is None or result["total_visits_left"] >= 0 else 0

        eligibility.surgery_amount_left = result["policy__product__max_amount_surgery"]
        eligibility.consultation_amount_left = result["policy__product__max_amount_consultation"]
        eligibility.delivery_amount_left = result["policy__product__max_amount_delivery"]
        eligibility.antenatal_amount_left = result["policy__product__max_amount_antenatal"]
        eligibility.hospitalization_amount_left = result["policy__product__max_amount_hospitalization"]

        if service:
            if service.category == Service.CATEGORY_SURGERY:
                if total_surgeries_left == 0 \
                        or services_left == 0 \
                        or (min_date_service and min_date_service > now) \
                        or (result["policy__product__max_amount_surgery"] is not None
                            and result["policy__product__max_amount_surgery"] <= 0):
                    eligibility.is_service_ok = False
                else:
                    eligibility.is_service_ok = True
            elif service.category == Service.CATEGORY_CONSULTATION:
                if total_consultations_left == 0 \
                        or services_left == 0 \
                        or (min_date_service and min_date_service > now) \
                        or (result["policy__product__max_amount_consultation"] is not None
                            and result["policy__product__max_amount_consultation"] <= 0):
                    eligibility.is_service_ok = False
                else:
                    eligibility.is_service_ok = True
            elif service.category == Service.CATEGORY_DELIVERY:
                if total_deliveries_left == 0 \
                        or services_left == 0 \
                        or (min_date_service and min_date_service > now) \
                        or (result["policy__product__max_amount_delivery"] is not None
                            and result["policy__product__max_amount_delivery"] <= 0):
                    eligibility.is_service_ok = False
                else:
                    eligibility.is_service_ok = True
            # Original code had a Service.CATEGORY_OTHER but with the same process as the else
            else:
                if services_left == 0 or (min_date_service and min_date_service > now):
                    eligibility.is_service_ok = False
                else:
                    eligibility.is_service_ok = True
        else:
            # It is a bit weird to return is_service_ok=True when there is no service but that's how it used to work
            eligibility.is_service_ok = True

        if items_left == 0 or (min_date_item and min_date_item > now):
            eligibility.is_item_ok = False
        else:
            eligibility.is_item_ok = True

        # The process above uses the None type but the stored procedure service sets these to 0
        eligibility.total_admissions_left = total_admissions_left
        eligibility.total_consultations_left = total_consultations_left
        eligibility.total_surgeries_left = total_surgeries_left
        eligibility.total_deliveries_left = total_deliveries_left
        eligibility.total_antenatal_left = total_antenatal_left
        eligibility.total_visits_left = total_visits_left
        return eligibility


def insert_renewals(date_from=None, date_to=None, officer_id=None, reminding_interval=None, location_id=None, location_levels=4):
    if reminding_interval is None:
        reminding_interval = PolicyConfig.policy_renewal_interval
    from core import datetime
    now = datetime.datetime.now()
    policies = Policy.objects.filter(
        status__in=[Policy.STATUS_EXPIRED, Policy.STATUS_ACTIVE],
        validity_to__isnull=True,
    )
    if reminding_interval:
        policies = policies.filter(expiry_date__lte=now + core.datetimedelta(days=reminding_interval))
    if location_id:
        # TODO support the various levels
        policies = policies.filter(
            Q(family__location_id=location_id)  # Village
            | Q(family__location__parent_id=location_id)  # Ward
            | Q(family__location__parent__parent_id=location_id)  # District
            | Q(family__location__parent__parent__parent_id=location_id)  # Region
        )
    if officer_id:
        policies = policies.filter(officer_id=officer_id)
    if date_from:
        policies = policies.filter(expiry_date__gte=date_from)
    if date_to:
        policies = policies.filter(expiry_date__lte=date_to)

    policies = policies.prefetch_related("product")

    for policy in policies:
        renewal_warning = 0
        renewal_date = policy.expiry_date + core.datetimedelta(days=1)
        product = policy.product  # will be updated if there is a conversion product
        officer = policy.officer
        # Get product code or substitution
        if not product.conversion_product_id:
            previous_products = []
            # Could also add a len(previous_products) < 20 but this avoids loops in the conversion_products
            while product not in previous_products and product.conversion_product:
                previous_products.append(product)
                product = product.conversion_product
            if product in previous_products:
                logger.error("The product %s has a substitution chain with a loop: %s, continuing with %s",
                             policy.product_id, [p.id for p in previous_products], product.id)

        # TODO allow this kind of comparison where the left side is a datetime
        # if datetime.datetime(product.date_from) <= renewal_date <= product.date_to:
        # noinspection PyChainedComparisons
        if renewal_date >= product.date_from and renewal_date <= product.date_to:
            renewal_warning |= 1

        # This is from the original code but is actually not possible as we have an inner join on it
        if not policy.officer_id:
            renewal_warning |= 2
        else:
            if officer:
                previous_officers = []
                while officer not in previous_officers and officer.substitution_officer:
                    previous_officers.append(officer)
                    officer = officer.substitution_officer
                if officer in previous_officers:
                    logger.error("The product %s has a substitution chain with a loop: %s, continuing with %s",
                                 policy.officer_id, [o.id for o in previous_officers], officer.id)
            if officer.works_to and renewal_date > officer.works_to:
                renewal_warning |= 4

        # Check if the policy has another following policy
        following_policies = Policy.objects.filter(family_id=policy.family_id) \
            .filter(Q(product_id=policy.product_id) | Q(product_id=product.id)) \
            .filter(start_date__gte=renewal_date)
        if not following_policies.first():
            policy_renewal, policy_renewal_created = PolicyRenewal.objects.get_or_create(
                policy=policy,
                validity_to=None,
                defaults=dict(
                    renewal_prompt_date=now,
                    renewal_date=renewal_date,
                    new_officer=officer,
                    phone_number=officer.phone,
                    sms_status=0,
                    insuree=policy.family.head_insuree,
                    policy=policy,
                    new_product=product,
                    renewal_warnings=renewal_warning,
                    validity_from=now,
                    audit_user_id=0,
                )
            )
            if policy_renewal_created:
                create_insuree_renewal_detail(policy_renewal)  # The insuree module can create additional renewal data


def update_renewals():
    from core import datetime
    now = datetime.datetime.now()
    updated_policies = Policy.objects.filter(validity_to__isnull=True, expiry_date__lt=now) \
        .update(status=Policy.STATUS_EXPIRED)
    logger.debug("update_renewals set %s policies to expired status", updated_policies)
    return updated_policies


@dataclass
class SmsQueueItem:
    index: int
    phone: str
    sms_message: str


def policy_renewal_sms(family_message_template, range_from=None, range_to=None, sms_header_template=None):
    if sms_header_template is None:
        sms_header_template = """--Renewal--
{{renewal.renewal_date}}
{{renewal.insuree.chf_id}}
{{renewal.last_name}} {{renewal.other_names}}
{{district_name|default_if_none:"district?"}}
{{ward_name|default_if_none:"ward?"}}
{{village_name|default_if_none:"village?"}}
{{renewal.new_product.code}}-{{renewal.new_product.name}}
{% for detail in renewal.details.all %}{% if detail.insuree.is_head_of_family %}
HOF{% endif %}
{{detail.insuree.chf_id}}
{{detail.insuree.last_name}} {{detail.insuree.other_names}}
{% endfor %}
"""
    sms_header = Template(sms_header_template)
    family_message = Template(family_message_template)
    from core import datetime
    now = datetime.datetime.now()
    sms_queue = []
    i_count = 0  # TODO: remove and make this method a generator

    if not range_from:
        range_from = now
    if not range_to:
        range_to = now

    renewals = PolicyRenewal.objects.filter(phone_number__isnull=False) \
        .filter(renewal_prompt_date__gte=range_from) \
        .filter(renewal_prompt_date__lte=range_to) \
        .prefetch_related("insuree") \
        .prefetch_related("new_officer") \
        .prefetch_related("new_product") \
        .prefetch_related("details") \
        .prefetch_related("details__insuree")

    for renewal in renewals:
        # Leaving the original code in comment to show it used to be handled. It is now delegated to the Django
        # template for the SMS, we just provide the list of renewal details
        #
        # first get the photo renewal string
        # for detail in renewal.details.filter(validity_to__isnull=True):
        #     if detail.insuree.chf_id == renewal.insuree.chf_id:  # not sure it's equivalent to checking the id
        #         head_photo_renewal = True
        #     else:
        #         sms_photos += f"\n{detail.insuree.chf_id}\n{detail.insuree.last_name} {detail.insuree.other_names}"
        # if len(sms_photos) > 0 or head_photo_renewal:
        #     head_text = "\nHOF" if head_photo_renewal else ""
        #     sms_photos += f"--Photos--{head_text}{sms_photos}" # added to sms_header

        village = renewal.policy.family.location
        sms_header_context = Context(dict(
            renewal=renewal,
            district_name=village.parent.parent if village and village.parent else None,
            ward_name=village.parent if village else None,
            village_name=village,
        ))
        sms_header_text = sms_header.render(sms_header_context)
        sms_message = sms_header_text

        if renewal.new_officer.phone_communication:
            sms_queue.append(SmsQueueItem(i_count, renewal.new_officer.phone, sms_message))
            i_count += 1

        # Create SMS for the family
        if family_message and renewal.insuree.phone:
            expiry_date = renewal.renewal_date - core.datetimedelta(days=1)
            new_family_message = family_message.render(Context(dict(
                insuree=renewal.insuree,
                renewal=renewal,
                expiry_date=expiry_date,
            )))
            # new_family_message = "" # REPLACE(REPLACE(REPLACE(REPLACE(REPLACE(REPLACE(@FamilyMessage, '@@InsuranceID', @CHFID), '@@LastName', @InsLastName),
            # '@@OtherNames', @InsOtherNames), '@@ProductCode', @ProductCode), '@@ProductName', @ProductName),
            # '@@ExpiryDate', FORMAT(@ExpiryDate,'dd MMM yyyy'))

            if new_family_message:
                sms_queue.append(SmsQueueItem(i_count, renewal.insuree.phone, new_family_message))
                i_count += 1

    return sms_queue


def update_insuree_policies(policy, audit_user_id):
    for member in policy.family.members.filter(validity_to__isnull=True):
        existing_ip = InsureePolicy.objects.filter(validity_to__isnull=True, insuree=member, policy=policy).first()
        if existing_ip:
            existing_ip.save_history()
        ip, ip_created = InsureePolicy.objects.filter(validity_to__isnull=True).update_or_create(
            insuree=member, policy=policy,
            defaults=dict(
                enrollment_date=policy.enroll_date,
                start_date=policy.start_date,
                effective_date=policy.effective_date,
                expiry_date=policy.expiry_date,
                offline=policy.offline,
                audit_user_id=audit_user_id
            )
        )
        if ip_created:
            logger.debug("Created InsureePolicy(%s) %s - %s", ip.id, member.chf_id, policy.uuid)
        else:
            logger.debug("Updated InsureePolicy(%s) %s - %s", ip.id, member.chf_id, policy.uuid)


def policy_status_premium_paid(policy, effective_date):
    if PolicyConfig.activation_option == PolicyConfig.ACTIVATION_OPTION_CONTRIBUTION:
        policy.effective_date = effective_date
        policy.status = Policy.STATUS_ACTIVE
    else:
        policy.status = Policy.STATUS_READY


def policy_status_payment_matched(policy):
    if PolicyConfig.activation_option == PolicyConfig.ACTIVATION_OPTION_PAYMENT \
            and policy.status == Policy.STATUS_IDLE:
        policy.status = Policy.STATUS_ACTIVE



# Generated by Django 3.2.18 on 2023-05-12 11:34

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('policy', '0005_add_foreign_keys'),
    ]

    operations = [
        migrations.AlterModelOptions(
            name='policy',
            options={'managed': True},
        ),
        migrations.AlterModelOptions(
            name='policyrenewal',
            options={'managed': True},
        ),
    ]